/*
 * Interface pour le programme de compression
 * (c) 1995 Fabrice Bellard
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "utypes.h"
#include "getopt.h"
#include "testcode.h"
#include "ppm.h"


/***************************************************************************
 * Interface avec les routines de compression
 */

#define METHOD_NB      2   /* nombre total de mthodes de compression */

#define METHOD_STORE   0
#define METHOD_PPM     1


#define DEFAULT_SUFFIX  ".st"    /* extension par dfault */
/* signature en dbut de fichier */
#define STAT_MAGIC_SIZE 4
char stat_magic[STAT_MAGIC_SIZE]={'S','T','F','B'};

FILE *my_file_in,*my_file_out;
int size_in,size_out;

UINT my_read_buf(UCHAR *buf,UINT len) {
	 UINT len1;
	 len1=fread(buf,1,len,my_file_in);
	 size_in+=len1;
	 return len1;
}

UINT my_write_buf(UCHAR *buf,UINT len) {
	 UINT len1;

	 len1=fwrite(buf,1,len,my_file_out);
	 size_out+=len1;
	 return len1;
}


/*
 * Compression de file_in vers file_out
 * Et impression ventuelle de quelques statistiques
 */
int File_Encode(FILE *file_in,FILE *file_out,int method,int verbose) {
	 clock_t timec;
	 float ratio;
	 int err;
	 
	 /* header */
	 fwrite(stat_magic,1,STAT_MAGIC_SIZE,file_out);
	 fputc(method,file_out);
	 
	 if (verbose)  {
			timec=clock();
	 }
	 size_in=0;
	 size_out=0;
	 my_file_in=file_in;
	 my_file_out=file_out;
	 
	 switch (method)  {
		case METHOD_STORE:
			err=CopyFile(my_read_buf,my_write_buf);
			break;

		case METHOD_PPM:
			err=PPM_EncodeFile(my_read_buf,my_write_buf);
			break;
			
		default:
			err=1;
			break;
	 }

	 if (verbose)  {
			timec=clock()-timec;
			if (timec<1) timec=1;
			fprintf(stderr,"In : %d bytes  Out: %d bytes\n", size_in,size_out);
			if (size_in!=0 && size_out!=0)  {
				 ratio=(float)size_out/(float)size_in;
				 fprintf(stderr,"Out/In: %0.2f%% (1:%0.1f)  Entropy: %0.2f bits/sym\n",
								ratio,1/ratio,8*ratio);
				 fprintf(stderr,"Time: %0.2f s  Speed: %0.1f kbytes/s\n", 
								(float) timec / CLOCKS_PER_SEC,
								(float) size_in * CLOCKS_PER_SEC / timec / 1024.0 );
			}
	 }

	 return err;
}
	 
/*
 * Dcompression
 */
int File_Decode(FILE *file_in,FILE *file_out,int verbose) {
	 int method,len;
	 char header[STAT_MAGIC_SIZE+1];
	 clock_t timec;
	 int err;
	 
	 /* lecture du header et controle du nombre "magique" */
	 
	 len=fread(header,1,STAT_MAGIC_SIZE+1,file_in);
	 if (strncmp(header,stat_magic,STAT_MAGIC_SIZE)!=0)  {
			return 1;
	 }
	 if (len!=(STAT_MAGIC_SIZE+1))  {
			return 1;
	 }
	 method=header[STAT_MAGIC_SIZE];

	 if (verbose)  {
			timec=clock();
	 }
	 size_in=0;
	 size_out=0;
	 my_file_in=file_in;
	 my_file_out=file_out;
	 switch(method)  {
		case METHOD_STORE:
			err=CopyFile(my_read_buf,my_write_buf);
			break;
		case METHOD_PPM:
			err=PPM_DecodeFile(my_read_buf,my_write_buf);
			break;
			
		default:
			err=1;
			break;
	 }

	 if (verbose)  {
			timec=clock()-timec;
			if (timec<1) timec=1;
			fprintf(stderr,"In : %d bytes  Out: %d bytes\n", size_in,size_out);
			if (size_in!=0 && size_out!=0)  {
				 fprintf(stderr,"Time: %0.2f s  Speed: %0.1f kbytes/s\n", 
								(float) timec / CLOCKS_PER_SEC,
								(float) size_out * CLOCKS_PER_SEC / timec / 1024.0 );
			}
	 }

	 return err;
}
	 


/*****************************************************************************
 * Traitement de la ligne de commande et interface utilisateur (rduite!)
 */

char *prog_name;          /* nom du programme */
int verbose;              /* vrai si en mode verbose */
int decompress;           /* vrai si dcompression */
int modepipe;             /* vrai si mode pipe */
int method;               /* mthode de compression */
int done;                 /* vrai si un fichier a t trait */
char *file_suffix;        /* suffixe pour les fichiers compresss */
int nodelete;             /* vrai si on n'efface pas le fichier  la fin */

/*
 * Traitement de flux
 */
int File_Doit1(FILE *file_in,FILE *file_out)  {
	 int err;
	 
	 if (decompress)  {
			err=File_Decode(file_in,file_out,verbose);
	 } else  {
			err=File_Encode(file_in,file_out,method,verbose);
	 }
	 if (err!=0)  {
			fprintf(stderr,"%s: encoding/decoding error\n",prog_name);
	 }
	 return err;
}
	 

/*
 * Traitement d'un fichier
 */

void File_Doit(char *filename_in)  {
	 char *filename_out;
	 FILE *file_in,*file_out;
	 int a,b;
	 
	 a=strlen(filename_in);
	 b=strlen(file_suffix);
	 filename_out=malloc(a+b+1);
	 
	 if (!decompress)  {
			strcpy(filename_out,filename_in);
			strcat(filename_out,file_suffix);
	 } else {
			if (a<=b || strcmp(filename_in+a-b,file_suffix)!=0)  {
				 fprintf(stderr,"%s: Bad Extension -- %s\n",prog_name,filename_in);
			} else  {
				 strcpy(filename_out,filename_in);
				 filename_out[a-b]='\0';
			}
	 }
			
	 if (verbose)  {
			if (decompress) fprintf(stderr,"Decompressing"); else fprintf(stderr,"Compressing");
			fprintf(stderr," '%s' to '%s'\n",filename_in,filename_out);
	 }
	 
	 file_in=fopen(filename_in,"r");
	 if (file_in==NULL)  {
			perror(filename_in);
	 } else  {
			file_out=fopen(filename_out,"w");
			if (file_out==NULL)  {
				 perror(filename_out);
			} else  {
				 if ( File_Doit1(file_in,file_out) != 0 )  {
						unlink(filename_out);
				 } else  {
						if (!nodelete) unlink(filename_in);
				 }
				 fclose(file_out);
			}
			fclose(file_in);
	 }
	 
	 free(filename_out);
}


/*
 * Aide 
 */

void Version(void) {
	 fprintf(stderr,"stat-1.00 (c) 1995 Fabrice Bellard\n");
}
	 
	 
void Help_Short(void)  {
	 fprintf(stderr,"Type %s -h for help.\n",prog_name);
}

void Help(void) {
	 Version();
	 fprintf(stderr,
"This is a *test* about PPM Compression
usage: %s [-hVvdpn] [-S suffix] [-a number] file...
-h (--help)        : This help
-V (--version)     : Print Version
-v (--verbose)     : Verbose mode (additive) (for testing only)
-d (--decompress)  : Decompress
-S .suf (--suffix) : Set suffix .suf on compressed files (now=%s)
-p (--pipe)        : Read from stdin & write to stdout
-a n (--algorithm) : Encoding method (0=Store, 1=PPM)
-n (--nodelete)    : Do not delete original file
"  
					 ,prog_name,file_suffix);
}


GETOPT_REC getopt_tab[]= {
	 {"help","h",0,'h'},
	 {"verbose","v",0,'v'},
	 {"memory","m",1,'m'},
	 {"decompress","d",0,'d'},
	 {"pipe","p",0,'p'},
	 {"version","V",0,'V'},
	 {"suffix","S",1,'S'},
	 {"algorithm","a",1,'a'},
	 {"nodelete","n",0,'n'},
	 {NULL,NULL,0,0}
};


void getopt_func(int option,GETOPT_REC *opt,char *param) {

	 switch (option)  {
		case GETOPT_UNKNOWN:
			fprintf(stderr,"%s: Unknown option: %s\n",prog_name,param);
			Help_Short();
			exit(1);
			
		case GETOPT_EXPPARAM:
			fprintf(stderr,"%s: Option '--%s' needs a parameter\n",prog_name,opt->long_name);
			Help_Short();
			exit(1);
			
		case GETOPT_NOPARAM:
			fprintf(stderr,"%s: Option '--%s' needs no parameter\n",prog_name,opt->long_name);
			Help_Short();
			exit(1);
		
		case GETOPT_ARG:
			done=1;
			File_Doit(param);
			break;

		case 'h':
			Help();
			exit(1);
		
		case 'V':
			Version();
			exit(1);

		case 'v':
			verbose++;
			break;

		case 'S':
			file_suffix=param;
			if (file_suffix==NULL || file_suffix[0]!='.')  {
				 fprintf(stderr,"%s: Bad suffix -- %s\n",prog_name,file_suffix);
				 exit(1);
			}
			break;
			
		case 'd':
			decompress=1;
			break;
			
		case 'p':
			done=1;
			File_Doit1(stdin,stdout);
			break;

		case 'a':
			method=atoi(param);
			if (method<0 || method>=METHOD_NB)  {
				 fprintf(stderr,"%s: Bad method -- %s\n",prog_name,param);
			}
			break;

		case 'n':
			nodelete=1;
			break;
			
		default:
		  if (param!=NULL) 
			 fprintf(stderr,"%s: unsupported option='--%s' param='%s'\n",
							 prog_name,opt->long_name,param);
			else
			 fprintf(stderr,"%s: unsupported option='--%s'\n",
							 prog_name,opt->long_name);
			break;
	 }
}



int main(int argc,char **argv)  {

 	 
	 prog_name=argv[0];
	 verbose=0;
	 decompress=0;
	 modepipe=0;
	 method=METHOD_PPM;
	 done=0;
	 file_suffix=DEFAULT_SUFFIX;
	 nodelete=0;
	 
	 if (argc<=1)  {
			Help();
			exit(1);
	 }
	 /* analyse des options et traitement associ */
	 GetOpt(argc,argv,getopt_tab,getopt_func);
	 
	 if (!done)  {
			fprintf(stderr,"%s: Nothing to do!\n",prog_name);
			exit(1);
	 }
	 
	 /* c'est fini :) */
	 
	 return 0;
}
	 
